-- Addon
local modName = ...;
local tb = CreateFrame("Frame",modName,UIParent);

-- Constants
local ITEM_SIZE = 36;
local PREVIEW_SIZE = 128;
local MAX_ENTRIES = 10;
local MAX_FILTERS = 4;

local TEXTURE_PATH = "Interface/Icons/";
local UNKNOWN_PATH = "|cffff6060Unknown icon path";

local MAX_ENUM_FEATS_OF_STRENGTH = 10000;
local MAX_ENUM_SPELLS = 550000;

-- Tables
local filterTags = {};			-- Tags used for filtering. Tags are seperated by underscore in the texture names. Only index 1, 2 and 3 are ever used.
local allTextures = {};			-- Array containing all textures
local displayList = {};			-- Array of which textures to show based on filters chosen. During OnShow: Counts textures when querying all textures, to avoid duplicate entries getting added to "allTextures".
local filterLists = {};			-- Used to build a list of valid filter strings for the dropdowns

tb.filterTags = filterTags;
tb.allTextures = allTextures;
tb.displayList = displayList;
tb.filterLists = filterLists;

local defConfig = {
};

--------------------------------------------------------------------------------------------------------
--                                          Helper Functions                                          --
--------------------------------------------------------------------------------------------------------

-- returns the path of a texture from the textureID -- sister function to GetFileIDFromPath()
local function GetTexturePathFromID(textureID)
	local path = ArtTexturePaths[textureID];
	if (not path) then
		--path = GetTextureInfo(btn.texture);
		--if (path ~= "string") then
		--	path = nil;
		--end
	end
	return path;
end

-- returns id and path no matter if the "texture" arg is id or path
local function GetTexturePathAndID(texture)
	if (type(texture) == "number") then
		return texture, GetTexturePathFromID(texture);
	elseif (type(texture) == "string") then
		return GetFileIDFromPath(texture), texture;
	end
end

-- Is texture valid given the current filter
local function IsTextureValid(texture,filterIndex)
	-- convert textureID into texture path
	if (type(texture) == "number") then
		texture = GetTexturePathFromID(texture);
		if (not texture) then
			return true;		-- if we cannot get the path, we cannot do filtering, so just include it
		end
	end

	-- ensure texture is a string
	assert(type(texture) == "string","Texture is not a string");

	-- strip parent path
	texture = texture:match("([^\\/]+)$");

	-- text filtering
	if (cfg.filter and not texture:upper():match(cfg.filter:upper())) then
		return false;
	end

	-- split path name into sections from the underscores, so we can apply filter to each section
	wipe(filterTags);
	local index = 1;
	for tag in texture:gmatch("[^_-]+") do
		filterTags[index] = tag:upper();
		index = (index + 1);
	end

	-- Does it pass the filters?
	for i = 1, filterIndex do
		local filter = cfg["filter"..i];
		if (filter and filter ~= "all" and filter ~= filterTags[i]) then
			return false;
		end
	end

	-- It's valid!
	return true;
end

-- Sort
local function SortDisplayListFunc(item1,item2)
	return item1:upper() < item2:upper();
end

-- First Time Initialize
local function FirstTimeInitialize()
	for i = 1, MAX_FILTERS do
		tb.filterDropDowns[i]:InitSelectedItem(cfg["filter"..i] or "all");
	end
	FirstTimeInitialize = nil;
end

--------------------------------------------------------------------------------------------------------
--                                          Initialize Frame                                          --
--------------------------------------------------------------------------------------------------------

-- OnShow
local function OnShow(self)
	self:EnumIconTexturesSimple();
--	self:EnumIconTextures();
--	self:EnumDiffs();

	if (FirstTimeInitialize) then
		FirstTimeInitialize();
	end

	self:BuildDisplayList();
end

-- OnHide -- Do an extreme cleanup here, we want to free up as many resources when done
local function OnHide(self)
	self.preview:SetTexture(nil);
	for i = 1, MAX_ENTRIES do
		local btn = tb.entries[i];
		btn.texture:SetTexture(nil);
		btn.path:SetText(nil);
	end
	wipe(allTextures);
	wipe(displayList);
	for i = 1, MAX_FILTERS do
		if (filterLists[i]) then
			wipe(filterLists[i]);
		end
	end
	wipe(filterLists);
	collectgarbage();
end

-- Event
function tb:VARIABLES_LOADED(event)
	if (not TextureBrowser_Config) then
		TextureBrowser_Config = {};
	end
	cfg = setmetatable(TextureBrowser_Config,{ __index = defConfig });
end
tb:RegisterEvent("VARIABLES_LOADED");

tb:SetSize(480,MAX_ENTRIES * ITEM_SIZE + 66 + 30);
tb:SetPoint("CENTER");
tb:SetBackdrop({ bgFile = "Interface\\ChatFrame\\ChatFrameBackground", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = { left = 3, right = 3, top = 3, bottom = 3 } });
tb:SetBackdropColor(0.1,0.22,0.35,1);
tb:SetBackdropBorderColor(0.1,0.1,0.1,1);
tb:EnableMouse(true);
tb:SetMovable(true);
tb:SetToplevel(true);
tb:SetClampedToScreen(true);
tb:Hide();

tb:SetScript("OnShow",OnShow);
tb:SetScript("OnHide",OnHide);
tb:SetScript("OnEvent",function(self,event,...) self[event](self,event,...) end);
tb:SetScript("OnMouseDown",tb.StartMoving);
tb:SetScript("OnMouseUp",tb.StopMovingOrSizing);

tb.outline = CreateFrame("Frame",nil,tb);
tb.outline:SetBackdrop({ bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = { left = 4, right = 4, top = 4, bottom = 4 } });
tb.outline:SetBackdropColor(0.1,0.1,0.2,1);
tb.outline:SetBackdropBorderColor(0.8,0.8,0.9,0.4);
tb.outline:SetPoint("TOPLEFT",12,-38);
tb.outline:SetPoint("BOTTOMRIGHT",-12,42);

tb.close = CreateFrame("Button",nil,tb,"UIPanelCloseButton");
tb.close:SetPoint("TOPRIGHT",-4,-4);
tb.close:SetScript("OnClick",function(self,button,down) tb:Hide(); end);

tb.header = tb:CreateFontString(nil,"ARTWORK","GameFontHighlight");
tb.header:SetPoint("TOPLEFT",12,-12);
tb.header:SetFont(tb.header:GetFont(),26,"THICKOUTLINE");

tb.preview = tb:CreateTexture("ARTWORK",nil);
tb.preview:SetPoint("TOPLEFT",tb,"TOPRIGHT",4,0)
tb.preview:SetWidth(PREVIEW_SIZE);
tb.preview:SetHeight(PREVIEW_SIZE);
tb.preview:SetTexCoord(0.07,0.93,0.07,0.93);

--------------------------------------------------------------------------------------------------------
--                                           Core Functions                                           --
--------------------------------------------------------------------------------------------------------

-- Add Texture -- Some icons appear more than once, this makes sure each unique icon is only added once to the full list
function tb:AddTexture(texture)
	if not (texture) then
		return;
	end
	--texture = texture:sub(TEXTURE_PATH:len() + 1,-1);	-- MoP: given texture paths no longer have "Interface\\Icons\\" prefixed.
--	local textureUpper = texture:upper();
	local textureUpper = texture;
	local count = (displayList[textureUpper] or 0);
	displayList[textureUpper] = (count + 1);
	if (count == 0) then
		allTextures[#allTextures + 1] = texture;
--	else
--		printf("[%.3d] %d",displayList[textureUpper],texture);
	end
end

function tb:EnumIconTexturesSimple()
	for id, path in next, ArtTexturePaths do
		if (path:match("Interface[\\/]Icons")) then
			self:AddTexture(path);
		end
	end
end

function tb:EnumDiffs()
	local f, n = 0, 0;
	-- 1
--	for index, id in ipairs(allTextures) do
--		if (not ArtTexturePaths[id]) then
--			n = n + 1;
--		else
--			f = f + 1;
--		end
--	end

	wipe(displayList);
	wipe(allTextures);

	tb:EnumIconTextures();
	local t1 = CopyTable(allTextures);		-- index, id

	wipe(displayList);
	wipe(allTextures);

	tb:EnumIconTexturesSimple();
	local t2 = CopyTable(allTextures);		-- index, path

	wipe(displayList);
	wipe(allTextures);

	for index, path in next, t2 do
		local id = GetFileIDFromPath(path);
		local idx = tIndexOf(t1,id);
		if (idx) then
			f = f + 1;
		else
			n = n + 1;
			print(path);
			tb:AddTexture(path);
		end
	end
	tb:BuildDisplayList();

	printf("found = %d, not found = %d, total = %d",f,n,#allTextures)
end

-- find as many valid icons as we can using the api
function tb:EnumIconTextures()
	-- FastCall
	local GetSpellInfo = _G.GetSpellInfo;
	local GetAchievementInfo = _G.GetAchievementInfo;
	local GetAchievementCategory = _G.GetAchievementCategory;

	local queryList = {};	-- Table passed to the icon query functions to be filled with valid texture names

	wipe(displayList);	-- Az: no need to call this here? already empty initially and cleared during OnHide()?

	-- Macro & Item Textures
	-- Az: MoP Use GetMacroIcons() and GetMacroItemIcons() now
	local addFromQueryFunction = function(func,list)
		func(wipe(list));
		for _, texture in ipairs(list) do
			self:AddTexture(texture);
		end
	end
	addFromQueryFunction(GetMacroIcons,queryList);
	addFromQueryFunction(GetMacroItemIcons,queryList);

	-- Achievement Textures -- Az: Achievement Statistic is not used, as it doesn't provide any additional textures
	for i = 1, GetCategoryNumAchievements(-1) do
		local _, _, _, _, _, _, _, _, _, icon = GetAchievementInfo(-1,i);
		if (icon and icon ~= "") then
			self:AddTexture(icon);
		end
	end

	-- Achievement Feats of Strength Textures
	for i = 1, MAX_ENUM_FEATS_OF_STRENGTH do
		local success, id, _, _, _, _, _, _, _, _, icon = pcall(GetAchievementInfo,i);
		if (success and id) and (GetAchievementCategory(id) == 81) then
			self:AddTexture(icon);
		end
	end

	-- Spells
	for i = 1, MAX_ENUM_SPELLS do
		local _, _, icon = GetSpellInfo(i);
		if (icon) then
			self:AddTexture(icon);
		end
	end
end

-- Build List, Sort and Update
function tb:BuildDisplayList()
	wipe(displayList);

	-- Check Textures
	for index, texture in ipairs(allTextures) do
		if (IsTextureValid(texture,MAX_FILTERS)) then
			displayList[#displayList + 1] = texture;
		end
	end
	wipe(filterTags);

	-- Sort & Update
	--sort(displayList,SortDisplayListFunc);
	tb.header:SetFormattedText("Texture Browser (|cffffff20%d|r)",#displayList);
	self.scroll:UpdateList();
end

--------------------------------------------------------------------------------------------------------
--                                           Entry Buttons                                            --
--------------------------------------------------------------------------------------------------------

tb.entries = {};

-- OnClick
local function Entry_OnClick(self,button)
	local editBox = ChatEdit_GetActiveWindow();
	if (IsModifiedClick("CHATLINK")) and (editBox) and (editBox:IsVisible()) then
		--editBox:Insert(self.texture:GetTexture():gsub("\\","\\\\"));
		local id, path = GetTexturePathAndID(displayList[self.index])
		editBox:Insert(path);
	end
end

-- OnEnter
local function Entry_OnEnter(self)
	tb.preview:SetTexture(self.texture:GetTexture());
end

-- OnLeave
local function Entry_OnLeave(self)
	tb.preview:SetTexture(nil);
end

-- Make Entries
for i = 1, MAX_ENTRIES do
	local btn = CreateFrame("Button",nil,tb);
	btn:SetHeight(ITEM_SIZE);

	btn:SetHighlightTexture("Interface\\QuestFrame\\UI-QuestLogTitleHighlight");
	btn:GetHighlightTexture():SetVertexColor(0.1,0.22,0.35);
	btn:SetScript("OnClick",Entry_OnClick);
	btn:SetScript("OnEnter",Entry_OnEnter);
	btn:SetScript("OnLeave",Entry_OnLeave);

	btn.texture = btn:CreateTexture(nil,"ARTWORK");
	btn.texture:SetPoint("TOPLEFT",2,-2);
	btn.texture:SetPoint("BOTTOMLEFT",2,2);
	btn.texture:SetWidth(ITEM_SIZE - 4);
	btn.texture:SetTexCoord(0.07,0.93,0.07,0.93);

	btn.path = btn:CreateFontString(nil,"ARTWORK","GameFontHighlight");
	btn.path:SetPoint("LEFT",btn.texture,"RIGHT",4,0);
	btn.path:SetTextColor(1,1,1);
	btn.path:SetJustifyH("LEFT");

	if (i == 1) then
		btn:SetPoint("TOPLEFT",18,-46);
		btn:SetPoint("TOPRIGHT",-12,-46);
	else
		btn:SetPoint("TOPLEFT",tb.entries[#tb.entries],"BOTTOMLEFT");
		btn:SetPoint("TOPRIGHT",tb.entries[#tb.entries],"BOTTOMRIGHT");
	end

	tb.entries[i] = btn;
end

-- Scroll
tb.scroll = CreateFrame("ScrollFrame",modName.."Scroll",tb,"FauxScrollFrameTemplate");
tb.scroll:SetPoint("TOPLEFT",tb.entries[1]);
tb.scroll:SetPoint("BOTTOMRIGHT",tb.entries[#tb.entries],-28,-1);
tb.scroll:SetScript("OnVerticalScroll",function(self,offset) FauxScrollFrame_OnVerticalScroll(self,offset,ITEM_SIZE,self.UpdateList); end);
--FauxScrollFrame_Update(tb.scroll,#tb.entries,MAX_ENTRIES,ITEM_SIZE);

-- Update List
function tb.scroll:UpdateList()
	FauxScrollFrame_Update(self,#displayList,MAX_ENTRIES,ITEM_SIZE);
	local index = self.offset;
	local mouseFocus = GetMouseFocus();
	for i = 1, MAX_ENTRIES do
		index = (index + 1);
		local btn = tb.entries[i];
		if (displayList[index]) then
			btn.index = index;

			--btn.texture:SetTexture(TEXTURE_PATH..displayList[index]);

			local id, path = GetTexturePathAndID(displayList[index]);
			path = path and path:gsub(TEXTURE_PATH,"") or UNKNOWN_PATH;
			btn.path:SetFormattedText("%s\n|cffffff80%d",path,id);

			btn.texture:SetTexture(id);

			if (btn == mouseFocus) then
				btn:GetScript("OnEnter")(btn);
			end

			btn:Show();
		else
			btn:Hide();
		end
	end
end

--------------------------------------------------------------------------------------------------------
--                                             UI Buttons                                             --
--------------------------------------------------------------------------------------------------------

do
	local function SetFilter(filterText)
		cfg.filter = filterText;
		tb:BuildDisplayList();
	end

	local uiButtons = {
		width = 60,
		height = 24,
		{
			text = "Filter",
			tip = "Set new filter\nShift-click to clear current filter",
			click = function(self)
				if (not IsShiftKeyDown()) then
					AzDialog:Show("Enter new filter...",cfg.filter or "",function(text) SetFilter(text); end)
				elseif (cfg.filter) then
					SetFilter(nil);
				end
			end,
		},
	}

	local function HideGTT(self)
		GameTooltip:Hide();
	end

	local function Button_OnEnter(self)
		if (self.tip) then
			GameTooltip:SetOwner(self,"ANCHOR_RIGHT");
			GameTooltip:SetText(self.tip);
		end
	end

	local lastButton;

	for _, btnDef in ipairs(uiButtons) do
		local b = CreateFrame("Button",nil,tb,"UIPanelButtonTemplate");

		b:SetText(btnDef.text);
		b:SetSize(btnDef.width or uiButtons.width or 72,btnDef.height or uiButtons.height or 24);

		if (lastButton) then
			b:SetPoint("RIGHT",lastButton,"LEFT",-4,0);
		else
			b:SetPoint("BOTTOMRIGHT",-12,14);
		end

		if (btnDef.disable) then
			b:Disable();
		end

		b:SetScript("OnClick",btnDef.click);
		b:SetScript("OnEnter",Button_OnEnter);
		b:SetScript("OnLeave",HideGTT);

		b.tip = btnDef.tip;

		local token = btnDef.text:gsub("%s","");
		tb["btn"..token] = b;
		lastButton = b;
	end
end

--------------------------------------------------------------------------------------------------------
--                                          Filter DropDowns                                          --
--------------------------------------------------------------------------------------------------------

tb.filterDropDowns = {};

do
	-- Builds the filter list for display in the dropdowns
	local function BuildDropdownFilterList(id)
		-- init the specific sub-table for this tag index
		if (not filterLists[id]) then
			filterLists[id] = {};
		end
		local list = filterLists[id];
		wipe(list);

		-- Check Textures
		for index, texture in ipairs(allTextures) do
			if (IsTextureValid(texture,id - 1)) then
				local tag = filterTags[id];
				if (tag) then
					if (tag:len() == 1) then
						tag = tag:upper();
					else
						tag = tag:match("(.)"):upper()..tag:match(".(.+)"):lower();
					end
					list[tag] = (list[tag] or 0) + 1;
				end
			end
		end

		-- Clear
		wipe(filterTags);
	end

	local function SortMenuItemsFunc(item1,item2)
		if (item1.value == "all") or (item2.value == "all") then
			return (item1.value == "all");
		else
			return item1.text < item2.text;
		end;
	end

	local function DropDown_Init(self,list)
		list[1].text = "|cff00ff00Show All";
		list[1].value = "all";
		BuildDropdownFilterList(self.id);
		local index = 2;
		for prefix, count in next, filterLists[self.id] do
			list[index].text = format("|cffffff00%s |cffc0c0c0(%d)|r",prefix,count);
			list[index].value = prefix:upper();
			index = (index + 1);
		end
		sort(list,SortMenuItemsFunc);
	end

	local function DropDown_SelectValue(self,entry,index)
		cfg["filter"..self.id] = entry.value;
		tb:BuildDisplayList();
	end

	local dropDownPadding = 4;
	local dropDownWidth = (tb.outline:GetWidth() - tb.btnFilter:GetWidth() + dropDownPadding * 0) / MAX_FILTERS - dropDownPadding;

	for i = 1, MAX_FILTERS do
		local dd = AzDropDown:CreateDropDown(tb,dropDownWidth,DropDown_Init,DropDown_SelectValue,true);

		dd:SetPoint("TOPLEFT",tb.outline,"BOTTOMLEFT",(i - 1) * (dropDownWidth + dropDownPadding),-4);
		dd.id = i;

		tb.filterDropDowns[i] = dd;
	end
end

--------------------------------------------------------------------------------------------------------
--                                           Slash Handling                                           --
--------------------------------------------------------------------------------------------------------

_G["SLASH_"..modName.."1"] = "/tb";
_G["SLASH_"..modName.."2"] = "/texturebrowser";
SlashCmdList[modName] = function(cmd)
	-- Show/Hide Dialog
	tb:SetShown(not tb:IsVisible());
end